function remoteConnection = getRemoteConnection(cluster)
%GETREMOTECONNECTION Get a connected RemoteClusterAccess
%
% getRemoteConnection will either retrieve a RemoteClusterAccess from the
% cluster's UserData or it will create a new RemoteClusterAccess.

% Copyright 2010-2021 The MathWorks, Inc.

% Store the current filename for the dctSchedulerMessages
currFilename = mfilename;

if ~isprop(cluster.AdditionalProperties, 'ClusterHost')
    error('parallelexamples:GenericSLURM:MissingAdditionalProperties', ...
        'Required field %s is missing from AdditionalProperties.', 'ClusterHost');
end
clusterHost = cluster.AdditionalProperties.ClusterHost;
if ~ischar(clusterHost)
    error('parallelexamples:GenericSLURM:IncorrectArguments', ...
        'Hostname must be a character vector');
end

if ~isprop(cluster.AdditionalProperties, 'RemoteJobStorageLocation')
    error('parallelexamples:GenericSLURM:MissingAdditionalProperties', ...
        'Required field %s is missing from AdditionalProperties.', 'RemoteJobStorageLocation');
end
remoteJobStorageLocation = cluster.AdditionalProperties.RemoteJobStorageLocation;
if ~ischar(remoteJobStorageLocation)
    error('parallelexamples:GenericSLURM:IncorrectArguments', ...
        'RemoteJobStorageLocation must be a character vector');
end
if isprop(cluster.AdditionalProperties, 'UseUniqueSubfolders')
    useUniqueSubfolders = cluster.AdditionalProperties.UseUniqueSubfolders;
    if ~islogical(useUniqueSubfolders)
        error('parallelexamples:GenericSLURM:IncorrectArguments', ...
            'UseUniqueSubfolders must be a logical scalar');
    end
else
    useUniqueSubfolders = false;
end

needToCreateNewConnection = false;
if isempty(cluster.UserData)
    needToCreateNewConnection = true;
else
    if ~isstruct(cluster.UserData)
        error('parallelexamples:GenericSLURM:IncorrectUserData', ...
            ['Failed to retrieve remote connection from cluster''s UserData.\n' ...
            'Expected cluster''s UserData to be a structure, but found %s'], ...
            class(cluster.UserData));
    end
    
    if isfield(cluster.UserData, 'RemoteConnection')
        % Get the remote connection out of the cluster user data
        remoteConnection = cluster.UserData.RemoteConnection;
        
        % And check it is of the type that we expect
        if isempty(remoteConnection)
            needToCreateNewConnection = true;
        else
            clusterAccessClassname = 'parallel.cluster.RemoteClusterAccess';
            if ~isa(remoteConnection, clusterAccessClassname)
                error('parallelexamples:GenericSLURM:IncorrectArguments', ...
                    ['Failed to retrieve remote connection from cluster''s UserData.\n' ...
                    'Expected the RemoteConnection field of the UserData to contain an object of type %s, but found %s.'], ...
                    clusterAccessClassname, class(remoteConnection));
            end
            
            if useUniqueSubfolders
                username = remoteConnection.Username;
                expectedRemoteJobStorageLocation = iBuildUniqueSubfolder(remoteJobStorageLocation, ...
                    username, iGetFileSeparator(cluster));
            else
                expectedRemoteJobStorageLocation = remoteJobStorageLocation;
            end
            
            if ~remoteConnection.IsConnected
                needToCreateNewConnection = true;
            elseif ~(strcmpi(remoteConnection.Hostname, clusterHost) && ...
                    remoteConnection.IsFileMirrorSupported && ...
                    strcmpi(remoteConnection.JobStorageLocation, expectedRemoteJobStorageLocation))
                % The connection stored in the user data does not match the cluster host
                % and remote location requested
                warning('parallelexamples:GenericSLURM:DifferentRemoteParameters', ...
                    ['The current cluster is already using cluster host %s and remote job storage location %s.\n', ...
                    'The existing connection to %s will be replaced.'], ...
                    remoteConnection.Hostname, remoteConnection.JobStorageLocation, remoteConnection.Hostname);
                cluster.UserData.RemoteConnection = [];
                needToCreateNewConnection = true;
            end
        end
    else
        needToCreateNewConnection = true;
    end
end

if ~needToCreateNewConnection
    return
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% CUSTOMIZATION MAY BE REQUIRED %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Get the credential options from the user using simple
% MATLAB dialogs or command line input.  You should change
% this section if you wish for users to provide their credential
% options in a different way.
% The pertinent options are:
% username - the username to use when running commands on the remote host
% useIdentityFile - whether or not to use an identity file (true/false).
%                   False means that a password is used
% identityFile - the full path to the identity file
% identityFileHasPassphrase - whether or not the identity file requires a passphrase
%                     (true/false).

% Use the UI if MATLAB has been started with a desktop
useUI = desktop('-inuse');
username = iGetUsername(cluster, useUI);

% Decide which authentication mode to use
% Default mechanism is to prompt for password
auth = 'Password';
if isprop(cluster.AdditionalProperties, 'AuthenticationMode')
    % If AdditionalProperties.AuthenticationMode is defined, use that
    auth = cluster.AdditionalProperties.AuthenticationMode;
    if ~ischar(auth) || ~ismember(auth, {'Agent', 'IdentityFile', 'Password'})
        error('parallelexamples:GenericSLURM:IncorrectArguments', ...
            'AuthenticationMode must be one of ''Agent'', ''IdentityFile'' or ''Password''');
    end
    if verLessThan('matlab', '9.11') && (~ischar(auth) || ~ismember(auth, {'IdentityFile', 'Password'}))
        error('parallelexamples:GenericSLURM:IncorrectArguments', ...
            'Prior to R2021b, AuthenticationMode must be one of ''IdentityFile'' or ''Password''');
    end
elseif isprop(cluster.AdditionalProperties, 'UseIdentityFile')
    % Otherwise use an identity file if UseIdentityFile is defined and true
    useIdentityFile = cluster.AdditionalProperties.UseIdentityFile;
    if ~islogical(useIdentityFile)
        error('parallelexamples:GenericSLURM:IncorrectArguments', ...
            'UseIdentityFile must be a logical scalar');
    end
    if useIdentityFile
        auth = 'IdentityFile';
    end
elseif isprop(cluster.AdditionalProperties, 'IdentityFile')
    % Otherwise use an identity file if IdentityFile is defined
    auth = 'IdentityFile';
else
    % Otherwise nothing is specified, ask the user what to do
    auth = iPromptUserForAuthenticationMode(cluster, useUI);
end
if verLessThan('matlab', '9.11')
    userArgs = {username};
else
    userArgs = {username, 'AuthenticationMode', auth};
end

% If using identity file, also need the filename and whether a passphrase is needed
if strcmp(auth, 'IdentityFile')
    identityFile = iGetIdentityFile(cluster, useUI);
    identityFileHasPassphrase = iGetIdentityFileHasPassphrase(cluster, useUI);
    userArgs = [userArgs, 'IdentityFilename', identityFile, ...
        'IdentityFileHasPassphrase', identityFileHasPassphrase];
    cluster.AdditionalProperties.UseIdentityFile = true;
end

cluster.saveProfile

% Now connect and store the connection
dctSchedulerMessage(1, '%s: Connecting to remote host %s', ...
    currFilename, clusterHost);
if useUniqueSubfolders
    remoteJobStorageLocation = iBuildUniqueSubfolder(remoteJobStorageLocation, ...
        username, iGetFileSeparator(cluster));
end
remoteConnection = parallel.cluster.RemoteClusterAccess.getConnectedAccessWithMirror(clusterHost, remoteJobStorageLocation, userArgs{:});
dctSchedulerMessage(5, '%s: Storing remote connection in cluster''s user data.', currFilename);
cluster.UserData.RemoteConnection = remoteConnection;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function username = iGetUsername(cluster, useUI)

if isprop(cluster.AdditionalProperties, 'Username')
    username = cluster.AdditionalProperties.Username;
    if ~ischar(username)
        error('parallelexamples:GenericSLURM:IncorrectArguments', ...
            'Username must be a character vector');
    end
    return
end

if useUI
    dlgMessage = sprintf('Enter the username for %s', cluster.AdditionalProperties.ClusterHost);
    dlgTitle = 'User Credentials';
    numlines = 1;
    usernameResponse = inputdlg(dlgMessage, dlgTitle, numlines);
    % Hitting cancel gives an empty cell array, but a user providing an empty string gives
    % a (non-empty) cell array containing an empty string
    if isempty(usernameResponse)
        % User hit cancel
        error('parallelexamples:GenericSLURM:UserCancelledOperation', ...
            'User cancelled operation.');
    end
    username = char(usernameResponse);
    cluster.AdditionalProperties.Username = username;
    return
end

% useUI == false
username = input(sprintf('Enter the username for %s:\n', cluster.AdditionalProperties.ClusterHost), 's');
cluster.AdditionalProperties.Username = username;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function identityFileHasPassphrase = iGetIdentityFileHasPassphrase(cluster, useUI)

if isprop(cluster.AdditionalProperties, 'IdentityFileHasPassphrase')
    identityFileHasPassphrase = cluster.AdditionalProperties.IdentityFileHasPassphrase;
    if ~islogical(identityFileHasPassphrase)
        error('parallelexamples:GenericSLURM:IncorrectArguments', ...
            'IdentityFileHasPassphrase must be a logical scalar');
    end
    return
end

if useUI
    dlgMessage = 'Does the identity file require a password?';
    dlgTitle = 'User Credentials';
    passphraseResponse = questdlg(dlgMessage, dlgTitle);
    if strcmp(passphraseResponse, 'Cancel')
        % User hit cancel
        error('parallelexamples:GenericSLURM:UserCancelledOperation', 'User cancelled operation.');
    end
    identityFileHasPassphrase = strcmp(passphraseResponse, 'Yes');
    cluster.AdditionalProperties.IdentityFileHasPassphrase = identityFileHasPassphrase;
    return
end

% useUI == false
validYesNoResponse = {'y', 'n'};
passphraseMessage = 'Does the identity file require a password? (y or n)\n';
passphraseResponse = iLoopUntilValidStringInput(passphraseMessage, validYesNoResponse);
identityFileHasPassphrase = strcmpi(passphraseResponse, 'y');
cluster.AdditionalProperties.IdentityFileHasPassphrase = identityFileHasPassphrase;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function identityFile = iGetIdentityFile(cluster, useUI)
if isprop(cluster.AdditionalProperties, 'IdentityFile')
    identityFile = cluster.AdditionalProperties.IdentityFile;
    if ~ischar(identityFile) || isempty(identityFile)
        error('parallelexamples:GenericSLURM:IncorrectArguments', ...
            'IdentityFile must be a nonempty character vector');
    end
else
    if useUI
        dlgMessage = 'Select Identity File to use';
        [filename, pathname] = uigetfile({'*.*', 'All Files (*.*)'},  dlgMessage);
        % If the user hit cancel, then filename and pathname will both be 0.
        if isequal(filename, 0) && isequal(pathname,0)
            error('parallelexamples:GenericSLURM:UserCancelledOperation', 'User cancelled operation.');
        end
        identityFile = fullfile(pathname, filename);
        cluster.AdditionalProperties.IdentityFile = identityFile;
    else
        identityFile = input(sprintf('Please enter the full path to the Identity File to use:\n'), 's');
        cluster.AdditionalProperties.IdentityFile = identityFile;
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function auth = iPromptUserForAuthenticationMode(cluster, useUI)
if useUI
    dlgMessage = sprintf('Use an identity file to login to %s?', cluster.AdditionalProperties.ClusterHost);
    dlgTitle = 'User Credentials';
    identityFileResponse = questdlg(dlgMessage, dlgTitle);
    if strcmp(identityFileResponse, 'Cancel')
        % User hit cancel
        error('parallelexamples:GenericSLURM:UserCancelledOperation', 'User cancelled operation.');
    end
    useIdentityFile = strcmp(identityFileResponse, 'Yes');
    cluster.AdditionalProperties.UseIdentityFile = useIdentityFile;
else
    validYesNoResponse = {'y', 'n'};
    identityFileMessage = sprintf('Use an identity file to login to %s? (y or n)\n', cluster.AdditionalProperties.ClusterHost);
    identityFileResponse = iLoopUntilValidStringInput(identityFileMessage, validYesNoResponse);
    useIdentityFile = strcmpi(identityFileResponse, 'y');
    cluster.AdditionalProperties.UseIdentityFile = useIdentityFile;
end

if useIdentityFile
    auth = 'IdentityFile';
else
    auth = 'Password';
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function returnValue = iLoopUntilValidStringInput(message, validValues)
% Function to loop until a valid response is obtained user input
returnValue = '';

while isempty(returnValue) || ~any(strcmpi(returnValue, validValues))
    returnValue = input(message, 's');
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function subfolder = iBuildUniqueSubfolder(remoteJobStorageLocation, username, fileSeparator)
% Function to build unique location using username and MATLAB release version
release = ['R' version('-release')];
subfolder = [remoteJobStorageLocation fileSeparator username fileSeparator release];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function fileSeparator = iGetFileSeparator(cluster)
% Function to return file separator for cluster operating system
if strcmpi(cluster.OperatingSystem, 'unix')
    fileSeparator = '/';
else
    fileSeparator = '\';
end
